/*
  JSDraw : the simple draw tool for web.
  (C) 2007 shin and sakurai

  JSDraw is freely distributable under the terms of an MIT-style license.
*/

/**
  JSONのユーティリティ
  復元機構を利用する場合は、 __type プロパティーを持っておく必要がある
*/
var UJSON = {};
/**
  JavaScriptオブジェクトをJSONテキストにする
*/
UJSON.obj2json = function(obj) {
	function object2json(obj) {
		var lines = [];
		for(var i in obj) {
			var val = obj[i];
			var type = typeof val;
			if (type == "function" || type == "undefined") {
				continue;
			}
			lines.push(i+": "+rec_obj2json(val));
		}
		return "{"+lines.join(",\n")+"}\n";
	}
	function array2json(arr) {
		var objarray = [];
		for(var i=0;i<arr.length;i++) {
			objarray[i] = rec_obj2json(arr[i]);
		}
		return "["+objarray.join(",")+"]\n";
	}
	function number2json(num) {
		return ""+num;
	}
	function bool2json(b) {
		return ""+b;
	}
	function string2json(s) {
		return "\""+s+"\"";
	}
	
	function rec_obj2json(obj) {
		var t = typeof obj;
		if (obj === null || t == "function" || t == "undefined") {
			return null;
		}
		switch(t) {
		case "number":
			return number2json(obj);
		case "boolean":
			return bool2json(obj);
		case "string":
			return string2json(obj);
		case "object":
			if (obj instanceof String) {
				return string2json(obj);
			}
			if (obj instanceof Array) {
				return array2json(obj);
			}
			return object2json(obj);
		}
		throw "Not supported object: "+obj;
	}
	return rec_obj2json(obj);
}
/**
  typeと復元関数のマップ（直接利用しない）
*/
UJSON.type2restore = {};
/**
  typeと復元関数の登録
*/
UJSON.registerType = function(typeName,restoreFunction) {
	this.type2restore[typeName] = restoreFunction;
}
/**
  typeプロパティーを持つデータからもとのオブジェクトを再現する
*/
UJSON.create = function(data) {
	if (data.__type) {
		return (this.type2restore[data.__type])(data);
	} else {
		alert("Can not restore an object: ["+data+"]");
	}
}

/**
  JSDraw本体
  
*/
function JSDraw(client,client_holder) {

	//############################################
	// 定数など
	//############################################

	var COLORS = ["black", "blue", "green", "orange", "red"];
	var LA = 5; //矢じりの大きさ

	//############################################
	// 描画ツール全体の状態
	//############################################

	var jsdraw = this;

	jsdraw.scroll = {
		scrollElement: null, //スクロールするDIVに入っているときに設定
		offsetTop: 0,
		offsetLeft: 0
	};
	
	//新規に追加される絵の設定。このあたりを部品で拡張すればリッチになる。
	jsdraw.currentPallet = {
		getLineColor: function() { return "red"; },
		getLineWidth: function() { return 2; },
		getInnerColor: function() { return null; },
		getFontColor: function() {return "red"; },
		getFontName: function() {return "SansSerif";},
		getFontSize: function() {return 14;}
	};

	//現在の描画ツールの状態
	jsdraw.context = {
		state: null,           // 現在の「状態受付オブジェクト」
		objects:  new Array(), // 各絵が入っているリスト
		selectedObject: null,  // 現在選択されているもの。無ければnull。

		pallet: jsdraw.currentPallet, // 上の新規絵用のパレット
		statusArea: null, // 現在の状態を表示するSPAN
		backgroundImage: null,

		jsonOutputArea: null,   //デバッグ用の出力

		deleteObject: function(obj) { // objects から引数のオブジェクトを消す
			obj.onDelete();
			this.objects = $A(this.objects).without(obj);
			if (this.selectedObject == obj) {
				selectedObject = null;
				transitState(new NormalState());
			}
			paint();
		}
	};

	//状態遷移を行う関数。必ずここで変更すること
	function transitState(nextState) {
		jsdraw.context.state = nextState;
		paint();
		outputJSONDebug();
	}

	function outputJSONDebug() {
		if (jsdraw.context.jsonOutputArea) {
			jsdraw.context.jsonOutputArea.value = jsdraw.toJSON();
		}
	}

	function getHitObject(objects,xx,yy) {
		for(var i=0,obj;obj = objects[i];i++) {
			if (obj.contains(xx,yy)) {
				return obj;
			}
		}
		return null;
	}

	function Rect(x,y,width,height) {
		var ex = x+width;
		var ey = y+height;
		this.x = Math.min(x,ex);
		this.ex = Math.max(x,ex);
		this.y = Math.min(y,ey);
		this.ey = Math.max(y,ey);
		this.width = Math.abs(width);
		this.height = Math.abs(height);
	}
	Rect.prototype.contains = function(x,y) {
		return (this.x <= x && x <= this.ex && 
				this.y <= y && y <= this.ey );
	}
	Rect.prototype.move = function(dx,dy) {
		this.x += dx; this.y += dy;
		this.ex += dx; this.ey += dy;
	}
	
	//############################################
	// 描画系
	//############################################
	
	function paint() {
		var context = jsdraw.context;
		var graphics = client.getContext("2d");
		graphics.globalCompositeOperation = 'source-over';

		//clear screen
		if(context.backgroundImage==null){
			graphics.beginPath();
			graphics.clearRect(0,0,client.offsetWidth,client.offsetHeight);
			graphics.closePath();
		} else {
			graphics.drawImage(context.backgroundImage, 0, 0,client.offsetWidth,client.offsetHeight);
		}

		//drawing objects loop
		var objects = context.objects;
		for (var i=0;i<objects.length;i++) {
			var obj = objects[i];
			graphics.save();
			obj.draw(graphics);
			graphics.restore();
		}
		
		//draw a selection
		if (context.selectedObject) {
			graphics.save();
			context.selectedObject.drawSelection(graphics);
			graphics.restore();
		}

		//draw custom image on the current state
		graphics.save();
		context.state.drawState(graphics);
		graphics.restore();
	}
	
	function drawOval(graphics,sx,sy,width,height,innerColor,lineColor){//###
		var g = graphics;

		//半径の縦の高さ(hh)と半径の横の長さ(hw)
		var hh = height/2;
		var hw = width/2;

		//中心座標(cx,cy)の計算
		var cx = sx + hw;
		var cy = sy + hh;

		var num = Math.min(Math.max(hh,hw)/2,30);
	
		if (num==0){
			return;
		}
		
		var dn = 2*Math.PI/num;

		//最初の位置へ
		g.beginPath();
		g.moveTo(cx+hw,cy);
		
		for(var i=-1; i<num ; i++){
			var kakudo = i * dn;

			//外の円の座標を求める
			var xi = cx + (hw * Math.cos(kakudo));
			var yi = cy + (hh * Math.sin(kakudo));

			g.lineTo(xi,yi);
		}
		g.stroke();
		g.closePath();
	}

	function drawLine(graphics,sx,sy,ex,ey,lineColor){//###
		var g = graphics;
		g.moveTo(sx,sy);
		g.lineTo(ex,ey);
		g.stroke();
		g.closePath();
	}

	function drawArrow(graphics,sx,sy,ex,ey,lineColor){//###
		var g = graphics;
		
		var dx=ex-sx;
		var dy=ey-sy;
		var length=Math.sqrt((dx*dx)+(dy*dy));

		if (length==0){
			return;
		}
		
		var cos = dx/length;
		var sin = -dy/length;
		
		var wkx1 = length-(Math.sqrt(3)*LA);
		var wky1 = 0-LA;
		var wkx2 = length-(Math.sqrt(3)*LA);
		var wky2 = 0+LA;

　　	var x1 = (wkx1*cos)+(wky1*sin);
		var y1 = (-wkx1*sin)+(wky1*cos);
	　　var x2 = (wkx2*cos)+(wky2*sin);
		var y2 = (-wkx2*sin)+(wky2*cos);

		g.beginPath();
		g.moveTo(sx,sy);
		g.lineTo(ex,ey);

		g.moveTo(ex,ey);
		g.lineTo(x1+sx,y1+sy);

		g.moveTo(ex,ey);
		g.lineTo(x2+sx,y2+sy);
		g.closePath();

		g.stroke();
	}

	//############################################
	// 入出力系
	//############################################
	
	jsdraw.toJSON = function() {
		var jos = [];
		var oos = jsdraw.context.objects;
		for(var i=0;i<oos.length;i++) {
			jos[i] = oos[i].toJSON();
		}
		return UJSON.obj2json({
			__type: "jsdraw", revision: "1", 
			objects: jos
		});
	}

	jsdraw.JSONto = function(text) {
		var oldData = jsdraw.context.objects;
		try {
			var source = "var data = "+text;
			eval(source);
			if (data.__type != "jsdraw") {
				alert("This is not jsdraw json format. ["+text+"]");
				return;
			}
			var newlist = [];
			var os = data.objects;
			for (var i=0;i<os.length;i++) {
				newlist.push(UJSON.create(os[i]));
			}
			//clear
			for (var i=0;i<oldData.length;i++) {
				oldData[i].onDelete();
			}
			jsdraw.context.objects = newlist;
		} catch (e) {
			if ( jsdraw.context.statusArea ) {
				jsdraw.context.statusArea.innerHTML = e.toString();
			}
			jsdraw.context.objects = oldData;
		}
		transitState(new NormalState());
	}

	function makePalletFromJSON(data) {
		return {
			getLineColor: function() { 
				return data.lineColor ? data.lineColor : null; },
			getLineWidth: function() { 
				return data.lineWidth ? data.lineWidth : null; },
			getInnerColor: function() { 
				return data.innerColor ? data.innerColor : null; },
			getFontColor: function() {
				return data.fontColor ? data.fontColor : null; },
			getFontName: function() {
				return data.fontName ? data.fontName : null;},
			getFontSize: function() {
				return data.fontSize ? data.fontSize : null;}
		};
	}

	//############################################
	// 図形オブジェクト系
	//############################################
	
	/**
       描画オブジェクトの共通抽象クラス
     */
	function AbstractDrawingObject(pallet) {
		if (pallet) {
			this.lineColor = pallet.getLineColor();
			this.lineWidth = pallet.getLineWidth();
			this.innerColor = pallet.getInnerColor();
		}
	}
	/**
       もし与えられた(x,y)でこのオブジェクトが応答する場合は true を返す。
     */
	AbstractDrawingObject.prototype.contains = function(x,y) {
		throw "Not implemented...";
	}
	/**
       移動するときに呼ばれる。引数は増分。
     */
	AbstractDrawingObject.prototype.move = function(dx,dy) {
		throw "Not implemented...";
	}
	AbstractDrawingObject.drawTransient = function(graphics,sx,sy,ex,ey) {//static method
		//do nothing
	}
	AbstractDrawingObject.prototype.draw = function(graphics) {
		//do nothing
	}
	AbstractDrawingObject.prototype.drawSelection = function(graphics) {
		//do nothing
	}
	AbstractDrawingObject.prototype.toJSON = function() {
		throw "Not implementation...";
	}
	/**
       削除されたときに呼ばれる。DOMから削除する場合など。
     */
	AbstractDrawingObject.prototype.onDelete = function() {
		//do nothing
	}

	function SquareObject(rect,pallet) {
		AbstractDrawingObject.call(this,pallet);
		this.x = rect.x;	this.y = rect.y;
		this.width = rect.width; this.height = rect.height;
	}
	SquareObject.prototype = new AbstractDrawingObject();
	SquareObject.prototype.getRect = function() {
		return new Rect(this.x,this.y,this.width,this.height);
	}
	SquareObject.prototype.contains = function(x,y) {
		return this.getRect().contains(x,y);
	}
	SquareObject.prototype.move = function(dx,dy) {
		this.x += dx; this.y += dy;
	}
	SquareObject.prototype.draw = function(graphics) {
		var g = graphics;
		g.lineWidth = this.lineWidth;
		if (this.innerColor) {
			g.fillStyle = this.innerColor;
			g.fillRect(this.x,this.y,this.width,this.height);
		}
		g.strokeStyle = this.lineColor;
		g.strokeRect(this.x,this.y,this.width,this.height);
	}
	SquareObject.prototype.drawSelection = function(graphics) {
		var g = graphics;
		var d = 3;
		g.strokeStyle = "orange";
		g.lineWidth = 2;
		g.shadowOffsetX = 4; g.shadowOffsetY = 4;
		g.shadowColor = "gray"; g.shadowBlur = 1;
		g.strokeRect(this.x-d,this.y-d,this.width+d*2,this.height+d*2);
	}
	SquareObject.drawTransient = function(graphics,sx,sy,ex,ey) {
		var g = graphics;
		g.strokeStyle = "gray";
		g.lineWidth = 1;
		g.strokeRect(sx,sy,ex-sx,ey-sy);
	}
	SquareObject.prototype.toJSON = function() {
		return {
			__type: "SquareObject",
			x: this.x, y: this.y, width: this.width, height: this.height,
			lineColor: this.lineColor, lineWidth: this.lineWidth, 
			innerColor: this.innerColor
		};
	}
	SquareObject.JSONto = function(data) {
		if (data.__type == "SquareObject") {
			return new SquareObject(new Rect(data.x, data.y, data.width, data.height),
									makePalletFromJSON(data));
		} else {
			alert("data type is not SquareObject: "+data.__type);
		}
	}
	UJSON.registerType("SquareObject",SquareObject.JSONto);

	function CircleObject(rect,pallet){
		AbstractDrawingObject.call(this,pallet);
		this.x = rect.x;
		this.y = rect.y;
		this.width = rect.width;
		this.height = rect.height;
	}
	CircleObject.prototype = new AbstractDrawingObject();
	CircleObject.prototype.getRect = function(){
		return new Rect(this.x,this.y,this.width,this.height)
	}

	CircleObject.prototype.contains = function(x,y){
		return this.getRect().contains(x,y);
	}
	CircleObject.prototype.move = function(dx,dy){
		this.x += dx;
		this.y += dy;
	}
	CircleObject.prototype.draw = function(graphics){
		var g = graphics;
		g.lineWidth = this.lineWidth;
		if(this.innnerColor){
			g.fillStyle = this.innerColor;
			drawOval(g,this.x, this.y, this.width, this.height);
		}
		g.strokeStyle = this.lineColor;
		drawOval(g,this.x, this.y, this.width, this.height);
	
	}
	CircleObject.prototype.drawSelection = function(graphics){
		var g = graphics;
		var d = 3;
		g.strokeStyle = "orange";
		g.linewidth = 2;
		g.shadowOffsetX = 4; g.shadowOffsetY = 4;
		g.shadowColor = "gray"; g.shadowBlur = 1;
		g.strokeRect(this.x-d,this.y-d,this.width+d*2,this.height+d*2);

	}
	CircleObject.drawTransient = function(graphics,sx,sy,ex,ey){
		var g = graphics;
		g.strokeStyle = "gray";
		g.lineWidth = 2;
		drawOval(g,sx,sy,ex-sx,ey-sy);
	}
	CircleObject.prototype.toJSON = function() {
		return {
			__type: "CircleObject",
			x: this.x, y: this.y, width: this.width, height: this.height,
			lineColor: this.lineColor, lineWidth: this.lineWidth, 
			innerColor: this.innerColor
		};
	}
	CircleObject.JSONto = function(data) {
		if (data.__type == "CircleObject") {
			return new CircleObject(new Rect(data.x, data.y, data.width, data.height),
									makePalletFromJSON(data));
		} else {
			alert("data type is not CircleObject: "+data.__type);
		}
	}
	UJSON.registerType("CircleObject",CircleObject.JSONto);

	//Line
	function LineObject(sx,sy,ex,ey,pallet){
		AbstractDrawingObject.call(this,pallet);
		this.sx = sx;
		this.sy = sy;
		this.ex = ex;
		this.ey = ey;
	}
	LineObject.prototype = new AbstractDrawingObject();
	LineObject.prototype.getRect = function(){
		return new Rect(this.sx,this.sy,this.ex-this.sx,this.ey-this.sy)
	}

	LineObject.prototype.contains = function(x,y){
		return this.getRect().contains(x,y);
	}
	LineObject.prototype.move = function(dx,dy){
		this.sx += dx;
		this.sy += dy;
		this.ex += dx;
		this.ey += dy;
	}
	LineObject.prototype.draw = function(graphics){
		var g = graphics;
		g.lineWidth = this.lineWidth;
		g.strokeStyle = this.lineColor;
		drawArrow(g,this.sx,this.sy,this.ex,this.ey);
	
	}
	LineObject.prototype.drawSelection = function(graphics){
		var g = graphics;
		var d = 3;
		g.strokeStyle = "orange";
		g.linewidth = 2;
		g.shadowOffsetX = 4; g.shadowOffsetY = 4;
		g.shadowColor = "gray"; g.shadowBlur = 1;
		var r = new Rect(this.sx, this.sy, this.ex - this.sx, this.ey - this.sy);
		g.strokeRect(r.x-d,r.y-d,r.width+d*2,r.height+d*2);
	}
	LineObject.drawTransient = function(graphics,sx,sy,ex,ey){
		var g = graphics;
		g.strokeStyle = "gray";
		g.lineWidth = 2;
		drawArrow(g,sx,sy,ex,ey);
	}
	LineObject.prototype.toJSON = function() {
		return {
			__type: "LineObject",
			sx: this.sx, sy: this.sy, ex: this.ex, ey: this.ey,
			lineColor: this.lineColor, lineWidth: this.lineWidth
		};
	}
	LineObject.JSONto = function(data) {
		if (data.__type == "LineObject") {
			return new LineObject(data.sx, data.sy, data.ex, data.ey,
								  makePalletFromJSON(data));
		} else {
			alert("data type is not LineObject: "+data.__type);
		}
	}
	UJSON.registerType("LineObject", LineObject.JSONto);

	//FreeLine
	function FreeLineObject(points,pallet){
		AbstractDrawingObject.call(this,pallet);
		this.points = points;
		//大きさ計測
		var sx = Infinity, sy = Infinity, ex = -Infinity, ey = -Infinity;
		for(var i=0,p; p=points[i]; i++) {
			sx = Math.min(p[0],sx);
			sy = Math.min(p[1],sy);
			ex = Math.max(p[0],ex);
			ey = Math.max(p[1],ey);
		}
		this.rect = new Rect(sx,sy,ex-sx,ey-sy);
	}
	FreeLineObject.prototype = new AbstractDrawingObject();
	FreeLineObject.prototype.getRect = function(){
		return this.rect;
	}
	FreeLineObject.prototype.contains = function(x,y){
		return this.getRect().contains(x,y);
	}
	FreeLineObject.prototype.move = function(dx,dy){
		for(var i=0,p; p=this.points[i]; i++) {
			p[0] += dx; p[1] += dy;
		}		
		this.rect.move(dx,dy);
	}
	FreeLineObject.prototype.draw = function(graphics){
		var g = graphics;
		g.lineWidth = this.lineWidth;
		g.strokeStyle = this.lineColor;
		g.moveTo(this.points[0][0],this.points[0][1]);
		for(var i=1,p; p=this.points[i]; i++) {
			g.lineTo(p[0],p[1]);
		}
		g.stroke();
	}
	FreeLineObject.prototype.drawSelection = function(graphics){
		var g = graphics;
		var d = 3;
		g.strokeStyle = "orange";
		g.linewidth = 2;
		g.shadowOffsetX = 4; g.shadowOffsetY = 4;
		g.shadowColor = "gray"; g.shadowBlur = 1;
		var r = this.getRect();
		g.strokeRect(r.x-d,r.y-d,r.width+d*2,r.height+d*2);
	}
	FreeLineObject.drawTransient = function(graphics,points){
		var g = graphics;
		g.strokeStyle = "gray";
		g.lineWidth = 2;
		g.moveTo(points[0][0],points[0][1]);
		for(var i=1,p; p=points[i]; i++) {
			g.lineTo(p[0],p[1]);
		}
		g.stroke();
	}
	FreeLineObject.prototype.toJSON = function() {
		return {
			__type: "FreeLineObject",
			points: this.points, 
			lineColor: this.lineColor, lineWidth: this.lineWidth
		};
	}
	FreeLineObject.JSONto = function(data) {
		if (data.__type == "FreeLineObject") {
			return new FreeLineObject(data.points,makePalletFromJSON(data));
		} else {
			alert("data type is not FreeLineObject: "+data.__type);
		}
	}
	UJSON.registerType("FreeLineObject",FreeLineObject.JSONto);

	//text box
	function TextObject(sx,sy,pallet,text) {
		AbstractDrawingObject.call(this,pallet);
		this.sx = sx;
		this.sy = sy;
		if (pallet) {
			this.fontname = pallet.getFontName();
			this.fontSize = pallet.getFontSize();
			this.fontColor = pallet.getFontColor();
		}

		//textbox
		this.object = document.createElement('DIV');
		this.object.style.position = 'absolute';
		this.object.style.fontSize = "18pt";
		this.object.style.color="red";
		this.object.type = "text";
		this.object.innerHTML= text ? text : "";
		this.setDIVPosition(sx,sy);

		client_holder.appendChild(this.object);
	}
	TextObject.prototype = new AbstractDrawingObject();
	TextObject.prototype.getRect = function(){
		return new Rect(this.sx,this.sy,
						this.object.offsetWidth,this.object.offsetHeight);
	}
	TextObject.prototype.setDIVPosition = function(x,y) {
		this.object.style.left = x+"px";
		this.object.style.top = y+"px";
	}
	TextObject.prototype.contains = function(x,y){
		return this.getRect().contains(x,y);
	}
	TextObject.prototype.move = function(dx,dy){
		this.sx += dx;
		this.sy += dy;
		this.setDIVPosition(this.sx,this.sy);
	}
	TextObject.prototype.draw = function(graphics){
		//ここでは何もしない
	}
	TextObject.prototype.drawSelection = function(graphics){
		var g = graphics;
		var d = 3;
		g.strokeStyle = "orange";
		g.linewidth = 2;
		g.shadowOffsetX = 4; g.shadowOffsetY = 4;
		g.shadowColor = "gray"; g.shadowBlur = 1;
		var r = this.getRect();
		g.strokeRect(r.x-d,r.y-d,r.width+d*2,r.height+d*2);
	}
	TextObject.drawTransient = function(graphics,sx,sy){
		//ここでは何もしない
	}
	TextObject.prototype.showEditor = function(stateObject) {
		//INPUTフィールドを表示して文字入力させる
		var e = document.createElement("INPUT");
		e.style.position = 'absolute';
		e.style.left = this.sx + "px";
		e.style.top = this.sy + "px";
		e.size = 30;
		e.value = this.object.innerHTML;
		var self = this;
		this.onBlur = function() {
			stateObject.onBlur();//フォーカスが外れたら確定にする
		};
		this.onKeydown = function(ev) {
			if (ev.keyCode == Event.KEY_RETURN) {
				stateObject.onBlur();//エンターも確定にする
			} else if (ev.keyCode == Event.KEY_ESC) {
				e.value = self.object.innerHTML;
				stateObject.onBlur();//ESCは内容を元に戻して確定にする
			}
		};
		Event.observe(e,'blur',this.onBlur,false);
		Event.observe(e,'keydown',this.onKeydown,false);
		client_holder.appendChild(e);
		setTimeout(function() {
			e.focus();//少したってからフォーカスを当てる
		},100);
		this.editor = e;
		this.object.style.display = "none";
	}
	TextObject.prototype.hideEditor = function(deleteFunction) {
		//このメソッドは TextEditingState からしか呼ばれない
		if (this.editor) {
			if (this.editor.value == "") {
				//空だったら削除する（図形はcontextにタッチしないため）
				deleteFunction();
			} else {
				this.object.innerHTML = this.editor.value;
				this.object.style.display = "";
			}
			Event.stopObserving(this.editor,'blur',this.onBlur,false);
			Event.stopObserving(this.editor,'keydown',this.onKeydown,false);
			client_holder.removeChild(this.editor);
		}
	}
	TextObject.prototype.onDelete = function() {
		client_holder.removeChild(this.object);
	}
	TextObject.prototype.toJSON = function() {
		return {
			__type: "TextObject",
			sx: this.sx, sy: this.sy, text: this.object.innerHTML,
			fontName: this.fontName, fontSize: this.fontSize, fontColor: this.fontColor
		};
	}
	TextObject.JSONto = function(data) {
		if (data.__type == "TextObject") {
			return new TextObject(data.sx,data.sy,makePalletFromJSON(data),data.text);
		} else {
			alert("data type is not TextObject: "+data.__type);
		}
	}
	UJSON.registerType("TextObject",TextObject.JSONto);

	//############################################
	// 状態オブジェクト
	//############################################
	
	// NomalState -> SelectedDownState
	// SelectedDownState -> SelectedNormalState, DragMovingState
	// SelectedNormalState -> SelectedDownState, NormalState
	// DragMovingState -> SelectedNormalState

	// SquareStartState -> SquareEndState
	// SquareEndState -> SelectedNormalState

	function AbstractState() {
	}
	AbstractState.prototype.debugState = function(text) {
		if (jsdraw.context && jsdraw.context.statusArea) {
			jsdraw.context.statusArea.innerHTML = text;
		}
	};
	AbstractState.prototype.onMouseDown = function(x,y) {};
	AbstractState.prototype.onMouseMove = function(x,y) {};
	AbstractState.prototype.onMouseUp = function(x,y) {};
	AbstractState.prototype.onDblClick = function(x,y) {};
	AbstractState.prototype.drawState = function(graphics) {};

	function NormalState() {
		this.debugState('normal');
		jsdraw.context.selectedObject = null;
	}
	NormalState.prototype = new AbstractState();
	NormalState.prototype.onMouseDown = function(xx,yy) {
		var obj = getHitObject(jsdraw.context.objects,xx,yy);
		if (obj) {
			transitState(new SelectedDownState(obj));
		}
	};

	function SelectedDownState(object) {
		this.debugState('selected down');
		jsdraw.context.selectedObject = object;
	}
	SelectedDownState.prototype = new AbstractState();
	SelectedDownState.prototype.onMouseUp = function(xx,yy) {
		transitState(new SelectedNormalState());
	}
	SelectedDownState.prototype.onMouseMove = function(xx,yy) {
		transitState(new DragMovingState(xx,yy));
	};
	
	function SelectedNormalState() {
		this.debugState('selected');
	}
	SelectedNormalState.prototype = new AbstractState();
	SelectedNormalState.prototype.onMouseDown = function(xx,yy) {
		var obj = getHitObject(jsdraw.context.objects,xx,yy);
		if (obj && obj == jsdraw.context.selectedObject) {
			transitState(new SelectedDownState(jsdraw.context.selectedObject));
		} else {
			if (obj) {
				transitState(new SelectedDownState(obj));
			} else {
				transitState(new NormalState());
			}
		}
	}
	SelectedNormalState.prototype.onDblClick = function(x,y) {
		var obj = getHitObject(jsdraw.context.objects,x,y);
		if (obj && (obj instanceof TextObject)) {
			transitState(new TextEditingState(obj));
		}
	};
	
	function DragMovingState(xx,yy) {
		this.debugState('drag moving');
		this.lastx = xx;
		this.lasty = yy;
	}
	DragMovingState.prototype = new AbstractState();
	DragMovingState.prototype.onMouseUp = function(xx,yy) {
		transitState(new SelectedNormalState());
	}
	DragMovingState.prototype.onMouseMove = function(xx,yy) {
		jsdraw.context.selectedObject.move(xx-this.lastx,yy-this.lasty);
		this.lastx = xx; this.lasty = yy;
		paint();
	};

	function SquareStartState() {
		this.debugState('square start');
		jsdraw.context.selectedObject = null;
	}
	SquareStartState.prototype = new AbstractState();
	SquareStartState.prototype.onMouseDown = function(x,y) {
		transitState(new SquareEndState(x,y));
	};

	function SquareEndState(sx,sy) {
		this.debugState('square end');
		this.sx = sx;
		this.sy = sy;
		this.ex = sx;
		this.ey = sy;
	}
	SquareEndState.prototype = new AbstractState();
	SquareEndState.prototype.drawState = function(graphics) {
		SquareObject.drawTransient(graphics,this.sx,this.sy,this.ex,this.ey);
	}
	SquareEndState.prototype.onMouseMove = function(ex,ey) {
		this.ex = ex;
		this.ey = ey;
		paint();
	};
	SquareEndState.prototype.onMouseUp = function(x,y) {
		var obj = new SquareObject(new Rect(this.sx,this.sy,this.ex-this.sx,
											this.ey-this.sy),jsdraw.context.pallet);
		jsdraw.context.objects.push(obj);
		jsdraw.context.selectedObject = obj;
		transitState(new SelectedNormalState());
	};


	//Circle
	function CircleStartState() {
		this.debugState('circle start');
		jsdraw.context.selectedObject = null;
	}
	CircleStartState.prototype = new AbstractState();
	CircleStartState.prototype.onMouseDown = function(x,y) {
		transitState(new CircleEndState(x,y));
	};

	function CircleEndState(sx,sy) {
		this.debugState('circle end');
		this.sx = sx;
		this.sy = sy;
		this.ex = sx;
		this.ey = sy;
	}
	CircleEndState.prototype = new AbstractState();
	CircleEndState.prototype.drawState = function(graphics) {
		CircleObject.drawTransient(graphics,this.sx,this.sy,this.ex,this.ey);
	}
	CircleEndState.prototype.onMouseMove = function(ex,ey) {
		this.ex = ex;
		this.ey = ey;
		paint();
	};
	CircleEndState.prototype.onMouseUp = function(x,y) {
		var obj = new CircleObject(new Rect(this.sx,this.sy,this.ex-this.sx,
											this.ey-this.sy),jsdraw.context.pallet);
		jsdraw.context.objects.push(obj);
		jsdraw.context.selectedObject = obj;
		transitState(new SelectedNormalState());
	};


	//Line
	function LineStartState() {
		this.debugState('line start');
		jsdraw.context.selectedObject = null;
	}
	LineStartState.prototype = new AbstractState();
	LineStartState.prototype.onMouseDown = function(x,y) {
		transitState(new LineEndState(x,y));
	};

	function LineEndState(sx,sy) {
		this.debugState('line end');
		this.sx = sx;
		this.sy = sy;
		this.ex = sx;
		this.ey = sy;
	}
	LineEndState.prototype = new AbstractState();
	LineEndState.prototype.drawState = function(graphics) {
		LineObject.drawTransient(graphics,this.sx,this.sy,this.ex,this.ey);
	}
	LineEndState.prototype.onMouseMove = function(ex,ey) {
		this.ex = ex;
		this.ey = ey;
		paint();
	};
	LineEndState.prototype.onMouseUp = function(x,y) {
		var obj = new LineObject(this.sx,this.sy,this.ex,this.ey,jsdraw.context.pallet);

		jsdraw.context.objects.push(obj);
		jsdraw.context.selectedObject = obj;
		transitState(new SelectedNormalState());
	};


	//FreeLine
	function FreeLineStartState() {
		this.debugState('free line start');
		jsdraw.context.selectedObject = null;
	}
	FreeLineStartState.prototype = new AbstractState();
	FreeLineStartState.prototype.onMouseDown = function(x,y) {
		transitState(new FreeLineEndState(x,y));
	};

	function FreeLineEndState(sx,sy) {
		this.debugState('free line drawing');
		this.points = [];
		this.points.push([sx,sy]);
	}
	FreeLineEndState.prototype = new AbstractState();
	FreeLineEndState.prototype.drawState = function(graphics) {
		FreeLineObject.drawTransient(graphics,this.points);
	}
	FreeLineEndState.prototype.onMouseMove = function(x,y) {
		this.points.push([x,y]);
		this.debugState('free line drawing['+this.points.length+']');
		paint();
	};
	FreeLineEndState.prototype.onMouseUp = function(x,y) {
		var obj = new FreeLineObject(this.points,jsdraw.context.pallet);

		jsdraw.context.objects.push(obj);
		jsdraw.context.selectedObject = obj;
		transitState(new SelectedNormalState());
	};


	//テキストボックス
	function TextPositioningState() {
		this.debugState('text position');
		jsdraw.context.selectedObject = null;
	}
	TextPositioningState.prototype = new AbstractState();
	TextPositioningState.prototype.onMouseDown = function(x,y) {
		var object = new TextObject(x,y,jsdraw.context.pallet);
		jsdraw.context.objects.push(object);
		jsdraw.context.selectedObject = object;
		transitState(new TextEditingState(object));
	};

	function TextEditingState(textObject) {
		this.debugState('text editing');
		textObject.showEditor(this);
		this.object = textObject;
		var self = this;
		this.deleteFunction = function() {
			jsdraw.context.deleteObject(self.object);
		};
	}
	TextEditingState.prototype = new AbstractState();
	TextEditingState.prototype.onBlur = function() {
		this.object.hideEditor(this.deleteFunction);
		jsdraw.context.selectedObject = this.object;
		transitState(new SelectedNormalState());
	}
	TextEditingState.prototype.onMouseDown = function(x,y) {
		var obj = getHitObject(jsdraw.context.objects,x,y);
		if (!obj || !(obj instanceof TextObject)) {
			this.object.hideEditor(this.deleteFunction);
			if (jsdraw.context.state instanceof TextEditingState) {
				//すでに別の状態でなければ通常状態に移行する
				transitState(new NormalState());
			}
		}
	}

	//############################################
	// イベント登録 
	//############################################
	
	transitState(new NormalState());

	// xxxx というマウスイベントを onXxxx というイベントハンドラに結びつける
	var eventTypes = "MouseDown MouseMove MouseUp DblClick".split(" ");
	for(var i=0,eventType;eventType = eventTypes[i];i++) {
		Event.observe(client_holder,
					  eventType.toLowerCase(),
					  (function(et) {
						  return function(ev) {
							  if (!jsdraw.context.state) return;
							  var x = Event.pointerX(ev) - client_holder.offsetLeft;
							  var y = Event.pointerY(ev) - client_holder.offsetTop;
							  var scr = jsdraw.scroll;
							  if (scr.scrollElement && scr.scrollElement.scrollTop) {
								  x += scr.scrollElement.scrollLeft;
								  y += scr.scrollElement.scrollTop;
							  }
							  x += scr.offsetLeft;
							  y += scr.offsetTop;
							  jsdraw.context.state[et].call(jsdraw.context.state,x,y);
/* for position debug
							  jsdraw.context.statusArea.innerHTML = 
								  "x:"+x+" y:"+y+
								  "  Ex:"+Event.pointerX(ev)+" Ey:"+Event.pointerY(ev)+
								  "  cx:"+client_holder.offsetLeft+" cy:"+client_holder.offsetTop;
*/
						  }
					  })("on"+eventType),false);
	}

	//############################################
	// 外部インタフェース接続
	//############################################

	this.setToolElements = function(params) {
		var buttons = "Square Circle Line FreeLine Text Delete".split(" ");
		for(var i=0,name;name=buttons[i];i++) {
			var paramName = name.toLowerCase()+"Button";
			if (paramName) {
				(this["set"+name+"Button"])(params[paramName]);
			}
		}
		
		if (params.statusArea) {
			this.setStatusArea(params.statusArea);
		}
		if (params.JSONOutput) {
			this.setJSONOutput(params.JSONOutput);
		}
		if (params.JSONInput) {
			this.setJSONInput(params.JSONInput);
		}
		if (params.scrollElement) {
			this.setScrollElement(params.scrollElement);
		}
	};

	this.setSquareButton = function(element) {
		Event.observe(element,'click',function() {
			transitState(new SquareStartState());
		},false);
	};

	this.setCircleButton = function(element) {
		Event.observe(element,'click',function() {
			transitState(new CircleStartState());
		},false);
	};

	this.setLineButton = function(element) {
		Event.observe(element,'click',function() {
			transitState(new LineStartState());
		},false);
	};

	this.setFreeLineButton = function(element) {
		Event.observe(element,'click',function() {
			transitState(new FreeLineStartState());
		},false);
	};

	this.setTextButton = function(element) {
		Event.observe(element,'click',function() {
			transitState(new TextPositioningState());
		},false);
	};

	this.setDeleteButton = function(element) {
		Event.observe(element,'click',function() {
			jsdraw.context.deleteObject(jsdraw.context.selectedObject);
		},false);
	};

	this.setColorSelection = function(element) {
		if (!element || element.tagName.toLowerCase() != "select") {
			return;
		}
	};

	this.setJSONOutput = function(element) {
		jsdraw.context.jsonOutputArea = element;
		outputJSONDebug();
	};

	this.setJSONInput = function(element) {
		jsdraw.JSONto(element.value);
	};

	this.setOffsetPosition = function(x,y) {
		jsdraw.scroll.offsetLeft = x;
		jsdraw.scroll.offsetTop = y;
	}

	this.setScrollElement = function(element) {
		jsdraw.scroll.scrollElement = element;
	};

	this.setStatusArea = function(element) {
		jsdraw.context.statusArea = element;
		transitState(jsdraw.context.state);
	};

	this.setBackgroundImage = function(img) {
		jsdraw.context.backgroundImage=img;
		paint();
	};

};
